/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "console.h"
#include "assert/assert.h"
#include "util/memory.h"
#include "util/string.h"

console_t CONSOLE;

void consoleInit() {
  memoryZero(&CONSOLE, sizeof(console_t));

  // Register the get and set command.
  CONSOLE.cmdGet = consoleRegCmd("get", consoleCmdGet);
  CONSOLE.cmdSet = consoleRegCmd("set", consoleCmdSet);
}

consolecmd_t * consoleRegCmd(const char_t *name, consolecmdfunc_t function) {
  consolecmd_t *cmd = &CONSOLE.commands[CONSOLE.commandCount++];
  consoleCmdInit(cmd, name, function);
  return cmd;
}

consolevar_t * consoleRegVar(
  const char_t *name,
  const char_t *value,
  consolevarchanged_t event
) {
  consolevar_t *var = &CONSOLE.variables[CONSOLE.variableCount++];
  consoleVarInitListener(var, name, value, event);
  return var;
}

void consolePrint(const char_t *message, ...) {
  char_t buffer[CONSOLE_LINE_MAX];
  va_list args;
  va_start(args, message);
  size_t len = vsnprintf(buffer, CONSOLE_LINE_MAX, message, args);
  va_end(args);

  assertTrue(len < CONSOLE_LINE_MAX, "Message is too long.");

  // Move all lines back
  memoryMove(
    CONSOLE.line[0],
    CONSOLE.line[1],
    (CONSOLE_HISTORY_MAX - 1) * CONSOLE_LINE_MAX
  );

  // Copy the new line
  memoryCopy(
    CONSOLE.line[CONSOLE_HISTORY_MAX - 1],
    buffer,
    len + 1
  );
  printf("%s\n", buffer);
}

void consoleExec(const char_t *line) {
  assertNotNull(line, "line must not be NULL");
  assertTrue(
    CONSOLE.execBufferCount < CONSOLE_EXEC_BUFFER_MAX,
    "Too many commands in the buffer."
  );

  if(line[0] == '\0') return;

  char_t buffer[CONSOLE_LINE_MAX + 1];
  size_t i = 0, j = 0;
  char_t c;

  do {
    c = line[i++];

    // Handle command separation by semicolon or end of line
    if(c == ';' || c == '\0') {
      // Null-terminate the current command and trim whitespace
      buffer[j] = '\0';
      stringTrim(buffer);

      // Skip empty commands
      if(buffer[0] == '\0') {
        j = 0;
        continue;
      }

      // Ensure the exec buffer is not full
      assertTrue(
        CONSOLE.execBufferCount < CONSOLE_EXEC_BUFFER_MAX,
        "Too many commands in the buffer."
      );

      // Create a new command execution
      consolecmdexec_t *exec = &CONSOLE.execBuffer[CONSOLE.execBufferCount++];
      memoryZero(exec, sizeof(consolecmdexec_t));

      // Parse command and arguments
      char_t *token = strtok(buffer, " ");
      while(token != NULL) {
        assertTrue(
          exec->argc < CONSOLE_CMD_ARGC_MAX,
          "Too many arguments in the command."
        );
        stringCopy(exec->argv[exec->argc++], token, CONSOLE_LINE_MAX);
        token = strtok(NULL, " ");
      }

      // First, see if there's a matching command.
      for(uint32_t k = 0; k < CONSOLE.commandCount; k++) {
        if(stringCompare(CONSOLE.commands[k].name, exec->argv[0]) != 0) {
          continue;
        }
        exec->cmd = &CONSOLE.commands[k];
      }

      // If match found continue
      if(exec->cmd != NULL) {
        j = 0;
        continue;
      }

      // If no match found, see if there's a matching variable.
      consolevar_t *var = NULL;

      for(uint32_t k = 0; k < CONSOLE.variableCount; k++) {
        if(stringCompare(CONSOLE.variables[k].name, exec->argv[0]) != 0) {
          continue;
        }

        var = &CONSOLE.variables[k];
        break;
      }

      // If no variable found, error out.
      if(!var) {
        consolePrint("Error: Command/Variable '%s' not found.", exec->argv[0]);
        CONSOLE.execBufferCount--;
        j = 0;
        continue;
      }

      // Found a matching variable, we basically change this entire command
      // to a get or a set. Arguments are reversed so they aren't stepped over
      if(exec->argc == 1) {
        exec->cmd = CONSOLE.cmdGet;
        exec->argc = 2;
        stringCopy(exec->argv[1], var->name, CONSOLE_LINE_MAX);
        stringCopy(exec->argv[0], CONSOLE.cmdGet->name, CONSOLE_LINE_MAX);
      } else {
        exec->cmd = CONSOLE.cmdSet;
        exec->argc = 3;
        stringCopy(exec->argv[2], exec->argv[1], CONSOLE_LINE_MAX);
        stringCopy(exec->argv[1], var->name, CONSOLE_LINE_MAX);
        stringCopy(exec->argv[0], CONSOLE.cmdSet->name, CONSOLE_LINE_MAX);
      }

      j = 0;
      continue;
    }

    if(j >= CONSOLE_LINE_MAX) {
      assertTrue(false, "Command exceeds maximum length.");
    }

    buffer[j++] = c; // Accumulate characters into the buffer
  } while(c != '\0');
}

void consoleProcess() {
  for(uint32_t i = 0; i < CONSOLE.execBufferCount; i++) {
    consolecmdexec_t *exec = &CONSOLE.execBuffer[i];

    assertTrue(
      exec->argc > 0,
      "Command execution has no arguments."
    );
    assertNotNull(
      exec->cmd,
      "Command execution has no command."
    );

    exec->cmd->function(exec);
  }

  // Clear the exec buffer
  CONSOLE.execBufferCount = 0;
}

void consoleCmdGet(const consolecmdexec_t *exec) {
  assertTrue(
    exec->argc >= 2,
    "get command requires 1 argument."
  );

  for(uint32_t i = 0; i < CONSOLE.variableCount; i++) {
    consolevar_t *var = &CONSOLE.variables[i];
    if(stringCompare(var->name, exec->argv[1]) != 0) continue;
    consolePrint("%s", var->value);
    return;
  }

  consolePrint("Error: Variable '%s' not found.", exec->argv[1]);
}

void consoleCmdSet(const consolecmdexec_t *exec) {
  assertTrue(
    exec->argc >= 3,
    "set command requires 2 arguments."
  );

  for(uint32_t i = 0; i < CONSOLE.variableCount; i++) {
    consolevar_t *var = &CONSOLE.variables[i];
    if(stringCompare(var->name, exec->argv[1]) != 0) continue;
    consoleVarSetValue(var, exec->argv[2]);
    return;
  }

  consolePrint("Error: Variable '%s' not found.", exec->argv[1]);
}